@@ -1,5 +1,5 @@
 """
- * $Id: 0.9+7 utils.py 25e913d62457 2020-04-03 od $
+ * $Id: 1.0+7 utils.py 37697a1cc48d 2020-07-08 od $
  *
  * This file is part of the Cloud Services Integration Platform (CSIP),
  * a Model-as-a-Service framework, API and application suite.
@@ -180,7 +180,7 @@ class Client(object):
             return c
         raise Exception("Illegal argument type: " + type(c))
 
-    def pubsub(self, url: str, callback=None) -> None:
+    def pubsub(self, url: str, callback=None, extra_metainfo: Dict = None) -> None:
         """Publish/subscribe execution """
 
         def cb(client: 'Client', file: str, counter: int):
@@ -191,25 +191,33 @@ class Client(object):
         if not hasattr(self, "webhook"):
             raise Exception("No webhook provided.")
         if not hasattr(self, "files"):
-            raise Exception("No files provided.")
+            raise Exception("No request files provided.")
 
-        h = {Client.HDR_CSIP_WEBHOOK: self.webhook}
+        # convert into tuple when only one webhook is given
+        if isinstance(self._webhook, str):
+            self._webhook = (self._webhook,)
 
+        webhook_count = len(self.webhook)
+
+        header = {}
         if hasattr(self, "bearertoken"):
-            h["Authorization"] = 'Bearer ' + self.bearertoken
+            header["Authorization"] = 'Bearer ' + self.bearertoken
 
         for i, file in enumerate(glob.iglob(self._files)):
-            if i + 1 < self._batch[0]:  # before min
-                continue
+            if i + 1 < self._batch[0]:
+                continue  # before min
             if i + 1 <= self._batch[1]:  # within range
-                h[Client.HDR_CSIP_REQUESTFILE] = Path(file).name
+                header[Client.HDR_CSIP_REQUESTFILE] = Path(file).name
+                header[Client.HDR_CSIP_WEBHOOK] = self.webhook[i % webhook_count]
                 c = Client.from_file(file)
-                resp = c.execute(url, True, None, h)
-                callback(resp, file, i)
+                if extra_metainfo is not None:
+                    c.add_metainfo(extra_metainfo)
+                resp = c.execute(url, True, None, header)
+                callback(resp, file, i + 1)
                 if self._delay > 0:
                     sleep(self._delay / 1000)
-            else:  # after max, we are done
-                break
+            else:
+                break  # after max, we are done
 
     def execute(self, url: str = None, sync: bool = True, files: List[str] = None, headers: Dict = None) -> 'Client':
         """Executes a service as HTTP/POST. This method returns when service is finished if sync
@@ -251,7 +259,7 @@ class Client(object):
         """Executes a service as HTTP/POST. This method runs the service asynchronously and
            returns when the service is done. It 'pings' the service run for completion in the
            frequency provided by 'fist_poll' and 'next_poll'. Provide a callback method to capture
-           the imtermittend query results.
+           the query results.
            
            Creates a reponse client.
            Use execute to call a service synchronously.
@@ -286,13 +294,13 @@ class Client(object):
 
         # Submitted -> Running
         r = resp.query_execute()
-        progress = r.get_progress();
+        progress = r.get_progress()
         callback(r, progress)
         while r.get_status() == Client.STATUS_RUNNING:
             sleep(np)
             # Running -> Running | Finished | Failed
             r = resp.query_execute()
-            progress = r.get_progress();
+            progress = r.get_progress()
             callback(r, progress)
         return r
 
@@ -371,11 +379,11 @@ class Client(object):
         self._bearertoken = token
 
     @property
-    def webhook(self) -> str:
+    def webhook(self) -> Tuple:
         return self._webhook
 
     @webhook.setter
-    def webhook(self, hook: str) -> None:
+    def webhook(self, hook: Tuple) -> None:
         self._webhook = hook
 
     @property
@@ -394,7 +402,7 @@ class Client(object):
         return self._batch
 
     @batch.setter
-    def batch(self, b: tuple) -> None:
+    def batch(self, b: Tuple) -> None:
         """For pubsub"""
         assert b[0] < b[1], "Invalid batch range"
         self._batch = b
@@ -419,6 +427,10 @@ class Client(object):
         """Set a metainfo entry"""
         self.metainfo[name] = value
 
+    def add_metainfo(self, d: Dict) -> None:
+        """Add a dict as metainfo"""
+        self.metainfo.update(d)
+
     def get_metainfo_names(self) -> List[str]:
         """Get all the entry names"""
         return list(self.metainfo.keys())
